#!/usr/bin/env python3

# Copyright (C) 2020-2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""In this module we have all the functions that have to do with dealing with
user-supplied arguments and finding the movie file.

The main function is :py:func:`~.get_args_movie`. This is function does all the
processing of the arguments, and imports the movie file. This function has the
non-trivial task of updating the argument list dynamically, depending on the
user-supplied options. For example, if a movie add custom command-line options, we must
display them in the help message.

A second function is :py:func:`~._init_argparse`. This function defines all the arguments
that can be customized via command-line. If you want to add new arguments, you have to
add them here. Notice, we do not use Python's argparse, but we use ``configargparse``.
This gives the user more flexibility on how to pass the arguments. Since there are tens
of options, users may want to have config files. Also, with ``configargparse`` some
options can be specified with a environment variable, which is nice for options like the
directory where to look for movies.

We also have :py:func:`~._is_movie_file`. This function parses a Python file and checks
if the requirements to be a "movie file" are met. These include containing a
``MOPIMovie`` class with specific methods. If new requirements arise, this function must
be edited.

The function :py:func:`~._check_args` checks the command-line arguments for possible
problems. It is wise to review this function if new options are added.

"""

import ast
import os
import sys
import warnings
from argparse import RawTextHelpFormatter

import configargparse


def _init_argparse(*args, **kwargs):
    """Initialize a new argparse parser and fill it with arguments.

    If you want to add new command-line options, here is where you have to look
    at.

    The arguments passed to this function and passed to the constructor of
    ``ArgParser``.

    All the arguments will be passed to the ``MOPIMovie`` class.

    :returns: Argparse parser with all the options except the custom ones added
              by the specific movie.
    :rtype: ``configargparse.ArgumentParser``

    """

    parser = configargparse.ArgParser(*args, **kwargs)

    general_options = parser.add_argument_group("General options")

    general_options.add(
        "movie",
        nargs="?",
        help="Movie to render among the ones found in MOPI_MOVIES_DIR. "
        "See bottom of the help message for list.",
    )
    general_options.add_argument("-m", "--movie-file", help="Path of the movie file.")
    general_options.add("-c", "--config", is_config_file=True, help="Config file path")
    general_options.add(
        "--movies-dir",
        default=".",
        help="Folder where to look form movies.",
        env_var="MOPI_MOVIES_DIR",
    )
    general_options.add_argument(
        "-o", "--outdir",
        default=".",
        help="Output directory for frames and video.",
    )
    general_options.add(
        "--disable-progress-bar",
        action="store_true",
        help="Do not display the progress bar when generating frames.",
    )
    general_options.add(
        "--parallel", help="Render frames in parallel.", action="store_true"
    )
    general_options.add(
        "--num-workers",
        default=os.cpu_count(),
        type=int,
        help="Number of cores to use (default: %(default)s)",
    )
    general_options.add(
        "-v", "--verbose", help="Enable verbose output", action="store_true"
    )
    general_options.add_argument(
        "-h",
        "--help",
        action="store_true",
        help="Show this help message and exit.",
    )

    video_options = parser.add_argument_group("Video rendering options")

    video_options.add_argument(
        "--movie-name",
        default="video",
        help="Name of output video file, without extension (default: %(default)s).",
    )
    video_options.add_argument(
        "--extension",
        default="mp4",
        help="File extension of the video (default: %(default)s).",
    )
    video_options.add_argument(
        "--fps",
        default="25",
        type=int,
        help="Frames-per-second of the video (default: %(default)s).",
    )
    video_options.add_argument("--author", help="Author metadata in the final video.")
    video_options.add_argument("--title", help="Title metadata in the final video.")
    video_options.add_argument("--comment", help="Comment metadata in the final video.")
    # If you add new metadata, you have to update the function
    # _metadata_from_args in moviemaker.py

    return parser


def _is_movie_file(path):
    """Check if the file path fulfills the minimum requirement to be used by
    ``motionpicture``.

    These are:
    - It must contain a class ``MOPIMovie``,
    - The class ``MOPIMovie`` must have a method ``__init__`` which has to take
      exactly one argument.
    - The class ``MOPIMovie`` must have a method ``self.get_frames``,
    - The class ``MOPIMovie`` must have a method ``self.make_frame`` which has
      to take exactly two arguments

    To do this, we open the file and inspect the content with ``ast``.

    If new requirement arise, you have to add them here!

    :param path: Path of the file to inspect.
    :type path: str

    :returns: Return True if all the requirements are met, False if at least one
              is not. Along with this, return a string with the description of
              the requirement that was not met (if all are met, the string will be
              empty).
    :rtype: tuple with bool and string

    """
    # We assume that path has been checked somewhere else.

    # First, we check if the file is Python file
    try:
        tree = ast.parse(open(path).read())
    except (SyntaxError, UnicodeDecodeError):
        return False, f"File {path} is not a valid Python file"

    # We check if there is a class ``MOPIMovie``.

    classes_in_tree = {node for node in tree.body if isinstance(node, ast.ClassDef)}

    movies = [class_ for class_ in classes_in_tree if class_.name == "MOPIMovie"]

    if len(movies) == 0:
        return False, f"File {path} does not contain a MOPIMovie class"

    # There must be one movie.
    movie_class = movies[0]

    methods_in_movie = {
        node for node in movie_class.body if isinstance(node, ast.FunctionDef)
    }

    init_method = [
        method for method in methods_in_movie if method.name == "__init__"
    ]

    if len(init_method) == 0:
        return (
            False,
            f"MOPIMovie class in file {path} does not contain a method __init__",
        )

    num_args_init = len(init_method[0].args.args)
    if num_args_init != 2:
        return (
            False,
            f"__init__ in MOPIMovie class in file {path} takes {num_args_init} arguments, "
            "it should take 2 (self, args)",
        )

    get_frames_method = [
        method for method in methods_in_movie if method.name == "get_frames"
    ]

    if len(get_frames_method) == 0:
        return (
            False,
            f"MOPIMovie class in file {path} does not contain a method get_frames",
        )

    make_frame_method = [
        method for method in methods_in_movie if method.name == "make_frame"
    ]

    if len(make_frame_method) == 0:
        return (
            False,
            f"MOPIMovie class in file {path} does not contain a method make_frame",
        )

    # There must be one make_frame.
    # It has to have exactly two arguments
    num_args = len(make_frame_method[0].args.args)
    if num_args != 3:
        return (
            False,
            f"make_framke in MOPIMovie class in file {path} takes {num_args} arguments, "
            "it should take 3 (self, path, frame_number)",
        )

    return True, ""


def _search_for_movies(folder):
    """Search in the given ``folder`` for files that satisfy the requirement
    specified by the function :py:func:`~._is_movie_file`.

    :param folder: Folder to scan.
    :type folder: str

    :returns: Set of files that contain movies.
    :rtype: set
    """
    files_in_folder = {
        file_
        for file_ in os.listdir(folder)
        if os.path.isfile(os.path.join(folder, file_))
    }
    return {
        file_
        for file_ in files_in_folder
        if _is_movie_file(os.path.join(folder, file_))[0]
    }


def _check_args(args):
    """Check if there are problems with the provided arguments.

    We check:
    - If the number of workers is larger than the number of CPUs.

    :param args: Arguments to check.
    :returns: Arguments as read from command line or from args
    :rtype: ``argparse.Namespace``

    """

    if args.num_workers > os.cpu_count():
        warnings.warn(
            f"You requested {args.num_workers} cores, "
            f"but the machine only has {os.cpu_count()}. "
            "This may result in performance loss."
        )


def get_args_movie(namespace, cli_args=None):
    """Process arguments and import `MOPIMovie` into `namespace``.

    When executing this function, you are also going to execute the movie file.

    This function calls :py:func:`~._init_argparse` to setup the options
    and :py:func:`~._check_args` to check for problems.

    :param namespace: Namespace in which to execute the movie file.
    :type namespace: dict

    :param args: List of arguments as if they were passed via command-line.
                 This is used only for testing.
    :type args: list

    :returns: Arguments as read from command line or from ``args``.
    :rtype: ``argparse.Namespace``

    """

    if cli_args is None:
        # Remove the name of the program from the list of arguments
        cli_args = sys.argv[1:]

    # We do a two-step parsing of the command-line arguments. With the first step,
    # we:
    # 1. Understand if a movie was selected
    # 2. Check what movies are available in the MOPI_MOVIES_DIR
    # 3. Update the help message adding choices for the possible movies.
    #
    # With the second step we actually parse all the arguments.
    #
    # The detailed strategy is the following:
    # 1. We start with a parser that has all the arguments common to all the movies.
    #    In this, we remove the help flag, as we added our own.
    # 2. We parse the arguments. This is first pass is to understand which movie
    #    the user has selected.
    # 3. We load the movie module, and add the arguments for that movie.
    # 4. We parse again the arguments.

    desc = (
        "Make a video specifying all the details using command-line arguments. "
        "To use this utility, you have to specify a movie. "
        "This code will look for movies in the MOPI_MOVIES_DIR, which you can customize. "
        "To select one of these movies, just pass the file name as first argument. "
        "Alternatively, you can pass the argument -m and specify a file."
    )

    parser = _init_argparse(
        description=desc, add_help=False, formatter_class=RawTextHelpFormatter
    )

    # We have to use this function because the user may have passed arguments
    # that are don't know yet (since we haven't loaded the movie's argument yet)
    parsed, unknown = parser.parse_known_args(cli_args)

    # Now we try to load the movie, if possible. If the user specified a
    # movie-file, that has to have the precedence.

    if parsed.movie_file:
        movie_file = parsed.movie_file
    elif parsed.movie:
        movie_file = os.path.join(parsed.movies_dir, parsed.movie)
    else:
        movie_file = None

    if movie_file:
        # Check if the specified movie file exists.
        if not (os.path.exists(movie_file) and os.path.isfile(movie_file)):
            raise ValueError(f"Movie-file {movie_file} does not exist.")

        # Check if the specified movie file meets the minimum requirements.
        is_movie_file, error_message = _is_movie_file(movie_file)
        if not is_movie_file:
            raise ValueError(error_message)

        # HACK: This is a nuclear way to achieve what we want.
        #
        # We want to import the user-supplied file. We could use importlib for
        # that. However, importing the movie as module introduces significant
        # complications in the parallelization. For instance, the imported file
        # is not in the PYTHONPATH of the various other sub-processes, which
        # makes impossible to pickle `MOPIMovie` (hence, it is impossible to use
        # multiprocessing!).
        #
        # The solution is to simply execute the file in the global scope. If we
        # do that, we can use multiprocessing without any problem.

        exec(open(movie_file).read(), namespace)

        # The user-supplied file can define a function `mopi_add_custom_options`
        # to add additional command-line options.
        if "mopi_add_custom_options" in namespace:
            custom_options = parser.add_argument_group("Movie custom options")
            namespace["mopi_add_custom_options"](custom_options)

    # Update the "help" adding the list of movies detected in the relevant folder.
    available_movies = _search_for_movies(parsed.movies_dir)

    if available_movies:
        epilog = f"Movies found in the MOPI_MOVIES_DIR ({parsed.movies_dir}):"
        for m in available_movies:
            epilog += f"\n* {m}"
    else:
        epilog = f"No movies found in the MOPI_MOVIES_DIR ({parsed.movies_dir})"

    parser.epilog = epilog

    if parsed.help:
        parser.print_help()
        sys.exit(0)

    if not movie_file:
        raise ValueError("Movie not specified. Please, specify a movie.")

    args = parser.parse_args(cli_args)

    _check_args(args)

    return args



#!/usr/bin/env python3

# Copyright (C) 2020 - 2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

import logging

from motionpicture.argparse import get_args_movie
from motionpicture import moviemaker as mm

def main():

    args = get_args_movie(globals())

    logger = logging.getLogger(__name__)

    if args.verbose:
        logging.basicConfig(format="%(asctime)s - %(message)s")
        logger.setLevel(logging.INFO)

    logger.info("Initializing MOPIMovie")
    # This is brought into the global namespace by get_args_movie.
    movie = MOPIMovie(args)
    logger.info("MOPIMovie initialized")

    logger.info("Getting frames")
    frames = movie.get_frames()
    logger.info("Frames gotten")

    frame_name_format = mm.prepare_frame_name_format(frames)
    logger.info(f"Chosen frame name format: {frame_name_format}")

    mm.check_outdir(args.outdir, frame_name_format)

    logger.info("Producing frames")
    mm.make_frames(
        movie,
        frames,
        args.outdir,
        frame_name_format,
        parallel=args.parallel,
        num_workers=args.num_workers,
        disable_progress_bar=args.disable_progress_bar,
    )
    logger.info("Frames produced")

    logger.info("Animating frames")
    mm.animate(
        args.movie_name,
        args.extension,
        args.outdir,
        frame_name_format,
        fps=args.fps,
        metadata=mm.metadata_from_args(args)
    )
    logger.info("Video rendered")

    print(f"Movie {args.movie_name} successfully created")

if __name__ == "__main__": # pragma: no cover

    main()



#!/usr/bin/env python3

# Copyright (C) 2020-2021 Gabriele Bozzola
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 3 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, see <https://www.gnu.org/licenses/>.

"""The :py:mod:`~.moviemaker` contains functions to drive the code that renders
the frames and to produce the final video.

One of the key functions is :py:func:`~.animate`, which takes the information on where
the frames are and how to produce the movie, and produces the movie with ffmpeg.

The second import function is :py:func:`~.make_frames`, which takes a ``MOPIMovie``
objects and renders all the required frames.

The additional functions provided are convenience functions to perform string
manipulation or error checking.

"""

import concurrent.futures
import os
from math import ceil, log10

import ffmpeg
from tqdm import tqdm


def check_outdir(output_folder, frame_name_format):
    """Check if the folder where frames should be created exists. If yes, check if
    has already some frames with the same extension has ``frame_name_format``. If yes,
    throw an error.

    (We don't support resuming the creation of frames yet).

    :param output_folder: File where frames have to be saved.
    :type output_folder: str
    :param frame_name_format: C-style format string for frame names.
    :type frame_name_format: str

    """
    if not os.path.isdir(output_folder):
        raise RuntimeError("{output_folder} does not exist")

    # Now we get the extension from frame_name_format
    extension = os.path.splitext(frame_name_format)[-1]

    # Now we find all the files with matching extension
    files_in_outdir_with_ext = {
        f for f in os.listdir(output_folder) if f.endswith(extension)
    }

    has_files = len(files_in_outdir_with_ext) > 0

    if has_files:
        raise RuntimeError(f"Directory {output_folder} already contains images")


def prepare_frame_name_format(frames, extension=".png"):
    """Take the list of frames that have to be generated, and compute the C-style format
    string necessary to accommodate all the frames.

    :param frames: List of frame identifiers.
    :type frames: list
    :param extension: File extension.
    :type extension: str

    :returns: C-style format string for frame names.
    :rtype str
    """
    number_of_digits = ceil(log10(len(frames)))

    extension = sanitize_file_extension(extension)

    return os.path.join(f"%0{number_of_digits}d{extension}")


def sanitize_file_extension(extension):
    """Return file extension that starts with '.' regardless of whether the given one
    had it or not.

    :param extension: File extension.
    :type extension: str

    :returns: File extension starting with '.'
    :rtype: str

    """

    if not extension.startswith("."):
        extension = f".{extension}"

    return extension


def make_frames(
    movie,
    frames,
    output_folder,
    frame_name_format,
    parallel=False,
    num_workers=None,
    disable_progress_bar=False,
):
    """Plot the frames, as directed by the given ``movie``. Optionally, plot frames in
    parallel.

    :param movie: Movie to use for rendering the frames.
    :type movie: ``MOPIMovie``
    :param frames: Which frames to render among the ones specified by
                   ``MOPIMovie.get_frames``.
    :type frames: list
    :param output_folder: Where to save the frames. The folder has to already exist.
                          Already existing frames with the same name will be overwritten.
    :type output_folder: str
    :param frame_name_format: C-style format string for the individual frames.
    :type frame_name_format: str
    :param parallel: If True, render multiple frames at the same time.
    :type parallel: bool
    :param num_workers: Number of frames rendered at the same time. If None, detect the
                        optimal number.
    :type num_workers: int or None
    :param disable_progress_bar: If True, do not display progress bar.
    :type disable_progress_bar: bool
    """
    frame_format_with_dir = os.path.join(output_folder, frame_name_format)

    # tqdm is the pretty progress bar, with estimate of remaining time.
    # It can be disabled passing disable_progress_bar=True
    with tqdm(total=len(frames), unit="frames", disable=disable_progress_bar) as pbar:
        if parallel:
            with concurrent.futures.ProcessPoolExecutor(
                max_workers=num_workers
            ) as executor:
                futures = {
                    executor.submit(
                        movie.make_frame, frame_format_with_dir % frame_num, frame
                    ): frame
                    for frame_num, frame in enumerate(frames)
                }
                for future in concurrent.futures.as_completed(futures):
                    frame_num = futures[future]

                    # We have to retrieve all the results to catch possible exceptions
                    try:
                        future.result()
                    except Exception as exc:
                        print(f"Frame {frame_num} generated an exception: {exc}")
                    pbar.update(1)
        else:
            for frame_num, frame in enumerate(frames):
                path = frame_format_with_dir % frame_num
                try:
                    movie.make_frame(path, frame)
                except Exception as exc:
                    print(f"Frame {frame_num} generated an exception: {exc}")
                pbar.update(1)


def metadata_from_args(args):
    """Extract the metadata fields from the command-line arguments and put them in a
    format ready to be fed to :py:func:`~_process_ffmpeg_metadata`.

    You have to update this when new metadata is added to the args!

    :param args: Command-line arguments.
    :type args: ``argparse.Namespace``

    :returns: Dictionary with metadata and the value.
    :rtype dict

    """
    # You have to update this when new metadata is added to the args!
    return {
        "artist": args.author,
        "comment": args.comment,
        "title": args.title,
    }


def _process_ffmpeg_metadata(metadata):
    """Process a dictionary of metadata in such a way that it can be interpreted by
    ffmpeg.

    :param metadata: Dictionary with keys the metadata fields, and values, the values of
                     the metadata.
    :type metadata: dict

    :returns: Metadata in a ffpmeg-compatible format.
    :rtype: dict

    """
    # Processing metadata following:
    # https://github.com/kkroening/ffmpeg-python/issues/112
    # Here we append the name of the metadata to make unique keys
    return {f"metadata:g:{k}": f"{k}={v}" for k, v in metadata.items()}


def animate(
    movie_name,
    movie_extension,
    output_folder,
    frame_name_format,
    fps=25,
    metadata=None,
    **kwargs,
):
    """Make a movie given the frames.

    Unknown arguments are passed to ffmpeg.

    :param movie_name: Name with extension of where to save the output within
                       ``output_folder``.
    :type movie_name: str
    :param movie_extension: File extension of the video.
    :type movie_extension: str
    :param frame_name_format: Format string for the names of the frames in
                              ``output_folder``.
    :type frame_name_format: str
    :param output_folder: Folder where the frames are saved and the video will be
                          produced.
    :type output_folder: str
    :param fps: Frames-per-second.
    :type fps: int
    :param metadata: Metadata to embed in the output file.
    :type metadata: dict or None

    """

    metadata = _process_ffmpeg_metadata(metadata) if metadata is not None else {}

    movie_file_name = os.path.join(
        output_folder, movie_name + sanitize_file_extension(movie_extension)
    )

    # Assemble movie with ffmpeg
    (
        ffmpeg.input(os.path.join(output_folder, frame_name_format))
        .filter("fps", fps=fps, round="up")
        .output(movie_file_name, **metadata, **kwargs)
        .overwrite_output()
        .run(quiet=True)
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['motionpicture']

package_data = \
{'': ['*']}

install_requires = \
['ConfigArgParse>=1.0.0,<2.0.0',
 'ffmpeg-python>=0.1.0,<0.2.0',
 'tqdm>=4.0.0,<5.0.0']

entry_points = \
{'console_scripts': ['mopi = motionpicture.mopi:main']}

setup_kwargs = {
    'name': 'motionpicture',
    'version': '0.1.2',
    'description': 'Make animations with Python',
    'long_description': '<p align="center">\n<img src="https://github.com/Sbozzolo/motionpicture/raw/master/logo.png" width="636" height="131">\n</p>\n\n[![GPLv3\nlicense](https://img.shields.io/badge/License-GPLv3-blue.svg)](http://perso.crans.org/besson/LICENSE.html)\n![Tests and documentation](https://github.com/Sbozzolo/motionpicture/workflows/Tests/badge.svg)\n[![codecov](https://codecov.io/gh/Sbozzolo/motionpicture/branch/master/graph/badge.svg?token=z7jvNNdwVS)](https://codecov.io/gh/Sbozzolo/motionpicture)\n\n# Introduction\n\n`motionpicture` is a Python library to simplify the creation of videos out of\nindividual frames. With `motionpicture`, you just have to specify how to produce\na generic frame, and the package will do everything else for you. In\n`motionpicture`, your code can be configured via command-line or text files:\nturning your code into a plug-in for `motionpicture` is trivial, so you will be\nable to reuse your code with ease.\n\n# Examples\n\nThere are two important ingredients to use `motionpicture`: `mopi`, and a\n_movie_ file. `mopi` is a command-line utility that comes when you install this\npackage. It will be your main interface to `motionpicture` and it has a\ncomprehensive `--help` function. A _movie_ file is a recipe on how to produce a\ngeneric frame. With few small restrictions, you have full control over this file\n(more info in section [Movie files](#movie-files)).\n\nIn these examples we are going to use `matplotlib` to do the plotting, but you\nare completely free to generate frames with any Python package you wish.\n\n## Unveiling a sine wave\n\nIn this example, we show how to use `mopi` to generate the following video.\n![sine_wave](https://github.com/Sbozzolo/motionpicture/raw/master/sine_wave.gif)\n\nTo produce this video, we need the following _movie_ file.\n``` python\nimport matplotlib.pyplot as plt\nimport numpy as np\n\nclass MOPIMovie:\n    def __init__(self, _args):\n        self.times = np.linspace(0, 10, 100)\n        self.values = np.sin(self.times)\n\n    def get_frames(self):\n        # Here we tell motionpicture what we consider a frame\n        return range(self.times)\n\n    def make_frame(self, path, frame_number):\n        # Here we plot a specific frame\n        plt.clf()\n        plt.plot(self.times[:frame_number], self.values[:frame_number])\n        plt.xlim([0, self.times[-1]])\n        plt.ylim([-1, 1])\n        plt.savefig(path)\n```\nAssuming this file is saved in `sin_wave.py`, we run\n\n``` sh\nmopi -m sin_wave.py -o frames_dir --parallel\n```\nThis is produce the individual frames in a folder `frames_dir` using all the CPUs\navailable on your machine. Then, it will glue the frames together in a video\nthat has the default name of `video.mp4`. If you want to change name, or other\nproperties (e.g., the fps), you can add options to `mopi`\n``` sh\nmopi -m sin_wave.py -o frames_dir --parallel --fps 10 --movie-name sin_wave\n```\nThis will produce a `sin_wave.mp4` video with 10 frames per second instead\n\n## Unveiling a sine wave with controllable frequency\n\nLet us continue on the example of the sine wave, and let us assume that we want\nto explore different frequencies.\n\nWe can edit the previous _movie_ file adding a `mopi_add_custom_options`\nfunction:\n\n``` python\ndef mopi_add_custom_options(parser):\n    """Add command-line options specific to this movie."""\n    parser.add_argument(\n        "-f",\n        "--frequency",\n        default=1,\n        type=int,\n        help="Frequency of the sine wave (default: %(default)s)",\n    )\n```\nThen, we edit the `__init__` function too:\n``` python\ndef __init__(self, args):\n    self.times = np.linspace(0, 10, 100)\n    self.values = np.sin(args.frequency * self.times)\n```\n_Movie_ files have to have an `__init__` that takes two arguments. The second\nis a `Namespace` that contains all the controllable options. These arguments\ncan be passed via command-line or configuration file.\n``` sh\nmopi -m sin_wave.py -o frames_dir --parallel --frequency 3\n```\nThis command will produce the following video.\n\n![sine_wave_fast](https://github.com/Sbozzolo/motionpicture/raw/master/sine_wave_fast.gif)\n\nAlternatively, you can put any of arguments in a config file `conf`, for example:\n\n``` text\noutdir: frames_dir\nfrequency: 3\n```\nConfig files support several syntaxes. Once you have the file, just call\n``` sh\nmopi sin_wave.py -c conf\n```\nYou can use config files and command-line options at the same time, but in case\nof conflict, the command-line arguments have the precedence.\n\n## Unveiling data in an arbitrary file\n\nNow that you have seen that you can control movies via command-line, it is time\nto introduce you to the plugin system in `motionpicture`.\n\nSuppose we have two-column files with time series data, we can modify the _movie_\nfile used in the previous example to animate those files, specifying which one\nat run-time.\n\n``` python\ndef mopi_add_custom_options(parser):\n    """Add command-line options specific to this movie."""\n    parser.add_argument(\n        "-f",\n        "--file",\n        required=True,\n        help="File to plot",\n    )\n```\nThen, we `import numpy as np` and edit the `__init__` function too:\n``` python\ndef __init__(self, args):\n    self.times, self.values = np.loadtxt(args.file).T\n    self.y_min, self.y_max = np.amin(self.value), np.amax(self.value)\n```\nWe computed the minimum and maximum of the value so that we can adjust the y axis\nrange. The  `make_frame` method will be the same, with the exception that we change\nthe `plt.ylim([-1, 1])` line to `plt.ylim([self.y_min, self.y_max])`.\n\nWe can save this file as `plot_timeseries` and call `mopi`:\n``` sh\nmopi -m plot_timeseries -o frames -f my_file.dat\n```\nOf course, we can add as many options as we wish to control the output. For instance,\nwe may want to add a switch to use logarithmic axes instead. The class\n`MOPIMovie` has full access to the user-supplied options, so you can do anything\nyou wish.\n\nWe did not hard-code anything in `plot_timeseries`, so the code will work for\nany dataset. However, if we want to use this file again, but in a different\nfolder, we would have to copy it over, since `mopi -m` expects the path of the\n_movie_ file. Alternatively, we can copy `plot_timeseries` to a specific folder\nof our choice, for example `~/.mopi_videos`. Then, we can set the environment\nvariable `MOPI_MOVIES_DIR` to be `~/.mopi_videos`, and `mopi` will be able to\nfind `plot_timeseries` from anywhere in your filesystem. In this case, you can\nsimply call:\n``` sh\nmopi plot_timeseries -o frames -f my_other_file.dat\n```\nEssentially, `plot_timeseries` became a plugin for `motionpicture` and you can\nanimate any data without having to write new code. This is one of the greatest\nstrengths of `motionpicture`, as it encourages you to write generic code that you\ncan easily reuse.\n\n# Installation\n\n`motionpicture` is available on PyPI. You can install it with `pip`:\n\n``` sh\npip3 install motionpicture\n```\n\nTo produce the final video, you have to have `ffmpeg` installed. Without\n`ffmpeg`, you will not be able to glue together the frames, but you can still\nuse `motionpicture` to render the frames.\n\n# _Movie_ files\n\nIn the language of `motionpicture`, a _movie_ file is a recipe on how to\ngenerate an individual frame. It is completely up to you how you do that, but\n`motionpicture` imposes some minimum requirements:\n\n- It has to be a valid Python 3 file\n- It has to contain a class `MOPIMovie` with a method `make_frame` and a method\n  `get_frames`.\n- The method `__init__` has to take two arguments.\n- The method `get_frames` has to return an iterable (e.g., a list) that\n  identifies each frame. The elements of this iterable are passed as the `frame`\n  argument to `make_frame`. \n- The method `make_frame` has to take two arguments, the `path` of the output of\n  the frame, and `frame`, the value that identifies frame (typically the frame\n  number). `path` is where the image has to be saved. You are in charge of\n  saving the image using the save method of your plotting package.\n\nOther than these requirements, you can do anything you want in the _movie_ file\n(e.g., you can add more methods, functions, classes...).\n\n> :warning: Due to its own nature, `motionpicture` has to execute any code that\n>           you supply. Do not use `motionpicture` with codes you do not trust!\n\n# Development\n\nWe use:\n* [Poetry](python-poetry.org) to manage dependencies, build, and publish\n  `motionpicture`.\n* [Black](https://github.com/psf/black) for formatting the code (with 89\n  columns).\n* [pytest](https://pytest.org) for unit tests (with `pytest-cov` for test\n  coverage).\n* GitHub actions for continuous integration.\n\nWe are happy to accept contributions.\n\n# Credits\n\nThe idea for `motionpicture` originated from the `SimVideo` package developed by\nWolfgang Kastaun.\n\n',
    'author': 'Gabriele Bozzola',
    'author_email': 'gabrielebozzola@arizona.edu',
    'maintainer': 'Gabriele Bozzola',
    'maintainer_email': 'gabrielebozzola@arizona.edu',
    'url': 'https://github.com/sbozzolo/motionpicture',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
